     /**
      * \brief Red Black Tree Implementation : Wrapper
      *
      * This module is a wrapper around the red black tree implemtree_rb_container_of_intentation
      * from the libtree project at github.com.
      * <p>
      * The wrapper adds type specific procedures for integer keys.
      * It also manages the dynamic memory allocation and deallocation.
      * The user can choose to make the allocation from a new or existing
      * user heap. The default heap can also be used.
      * <p>
      * The tree has no maximum number of elements. As long as you can get
      * memory from your heap you can add elements to the tree.
      *
      * \author Mihael Schmidt
      * \date   2011-02-07
      *
      * \link https://github.com/fbuihuu/libtree libtree project
      * \link http://www.rpgnextgen.com RPG Next Gen
      *
      * \todo procedures for string keys
      */


     H nomain


      *-------------------------------------------------------------------------
      * Prototypes
      *-------------------------------------------------------------------------
     D countNode       PR
     D   node                          *   const
     D   count                       10I 0

     D removeNodesRecursively...
     D                 PR
     D   tree                          *   const
     D   node                          *

      /include 'libtree_h.rpgle'
      /include 'ceeapi_h.rpgle'
      /include 'libc_h.rpgle'


      *-------------------------------------------------------------------------
      * Data Structures
      *-------------------------------------------------------------------------
     D tree_header_t   DS                  qualified template align
     D   heapId                      10I 0
     D   tree                          *
     D   heapCreated                   N
     D   deallocate                    N
      *
     D tree_rb_node_int_t...
     D                 DS                  qualified template align
     D  key                          10I 0
     D  value                          *
     D  length                       10I 0
     D  node                               likeds(rbnode_t)
      

      *-------------------------------------------------------------------------
      * Procedures
      *-------------------------------------------------------------------------

     /**
      * \brief Create tree
      *
      * Creates a new tree. If no id for a heap (user or default heap) is
      * provided a new user heap will be created for memory allocation.
      * <p>
      * A compare function must be passed to the constructor for comparing two
      * nodes. See <code>tree_rb_int_compare</code> for an implementation
      * with integer keys.
      * <p>
      * Per default any memory from removed nodes will be deallocated but you
      * can also passed *on as the third parameter and have the system free all
      * allocated resources on activation group end.
      * 
      * \param Compare function
      * \param Heap id (default: create new heap)
      * \param Deallocate memory (default: *on)
      *
      * \return Pointer to new tree
      *
      * \info The calling program must make sure to free the memory of the tree
      *       either with the procedure <code>tree_rb_finalize</code> or by
      *       discarding the heap manually. All allocated memory will automatically
      *       be free with the end of the activation group.
      */
     P tree_rb_create  B                   export
     D                 PI              *
     D   compareFunction...
     D                                 *   procptr value
     D   pHeapId                     10I 0 const options(*nopass : *omit)
     D   pDeallocate                   N   const options(*nopass)
      *
     D header          DS                  likeds(tree_header_t) based(tree)
     D heapId          S             10I 0
     D heapCreated     S               N   
     D deallocate      S               N   inz(*on)
      /free
       if (%parms() >= 2 and %addr(pHeapId) <> *null);
         heapId = pHeapId;
       else;
         cee_createHeap(heapId : *omit : *omit : *omit : *omit);
         heapCreated = *on;
       endif;

       if (%parms() = 3);
         deallocate = pDeallocate;
       endif;

       cee_getStorage(heapId : %size(tree_header_t) : tree : *omit);
       cee_getStorage(heapId : %size(rbtree_t) : header.tree : *omit);

       rbtree_init(header.tree : compareFunction);
       header.heapId = heapId;
       header.heapCreated = heapCreated;
       header.deallocate = deallocate;
       
       return tree;
      /end-free
     P                 E


     /**
      * \brief Dispose tree
      *
      * Frees all allocated resources if the tree uses its own heap. If the
      * user has passed an heap id to an existing heap on tree creation then
      * the user must take of freeing the allocated resources (f. e. by
      * clearing the tree or by discarding the heap).
      *
      * \param Pointer to tree
      */
     P tree_rb_finalize...
     P                 B                   export
     D                 PI
     D  tree                           *
      *
     D header          DS                  likeds(tree_header_t) based(tree)
      /free
       if (header.heapCreated);
         cee_discardHeap(header.heapId : *omit);
       endif;
      /end-free
     P                 E


     /**
      * \brief Insert new element
      *
      * Inserts a new element into the tree. Any existing element will be
      * replaced. The resources of any existing element will be freed.
      * <p>
      * The value is a passed as a pointer to the real value. If the length
      * parameter is omitted it is assumed that the value is null-terminated.
      * <p>
      * The key and value will be copied to the tree node.
      *
      * \param Pointer to tree
      * \param Key
      * \param Pointer to value
      * \param Value length (default: value must be null-terminated)
      */
     P tree_rb_int_put...
     P                 B                   export
     D                 PI
     D  tree                           *   const
     D  key                          10I 0 const
     D  value                          *   const options(*string)
     D  length                       10I 0 const options(*nopass)
      *
     D header          DS                  likeds(tree_header_t) based(tree)
     D node            DS                  likeds(tree_rb_node_int_t)
     D                                     based(nodePtr)
     D null            S              1A   inz(x'00')
     D embeddedPtr     S               *
     D userNodePtr     S               *
      /free
       cee_getStorage(header.heapId : %size(tree_rb_node_int_t) :
                      nodePtr : *omit);
       node.key = key;

       if (%parms() = 3);
         node.length = strlen(value);
         cee_getStorage(header.heapId : node.length + 1 : node.value : *omit);
         memcpy(node.value : value : node.length + 1); // also copy the null value at the end
       else;
         node.length = length;
         cee_getStorage(header.heapId : node.length + 1 : node.value : *omit);
         memcpy(node.value : value : node.length);
         // add null
         memcpy(node.value + node.length : %addr(null) : 1);
       endif;

       embeddedPtr = rbtree_insert(%addr(node.node) : header.tree);
       if (embeddedPtr <> *null); // means the key does already exist in the tree
         // remove existing entry
         userNodePtr = tree_rb_int_container_of(embeddedPtr);
         tree_rb_int_removeByNode(tree : userNodePtr);

         // add new entry
         rbtree_insert(%addr(node.node) : header.tree);
       endif;
      /end-free
     P                 E


     /**
      * \brief Get element
      *
      * Returns a pointer to the user node of the given key.
      *
      * \param Pointer to tree
      * \param Key
      *
      * \return Pointer to user data structure or *null if there is no element
      *         with the given key in the tree.
      */
     P tree_rb_int_get...
     P                 B                   export
     D                 PI              *
     D   tree                          *   const
     D   key                         10I 0 const
      *
     D header          DS                  likeds(tree_header_t) based(tree)
     D node            DS                  likeds(tree_rb_node_int_t)
     D embeddedPtr     S               *
      /free
       clear node;
       node.key = key;
       embeddedPtr = rbtree_lookup(%addr(node.node) : header.tree);
       return tree_rb_int_container_of(embeddedPtr);
      /end-free
     P                 E


     /**
      * \brief Get first element
      *
      * Returns the first element of the tree.
      *
      * \param Pointer to tree
      *
      * \return Pointer to user data structure or *null if the tree is empty.
      */
     P tree_rb_int_first...
     P                 B                   export
     D                 PI              *
     D   tree                          *   const
      *
     D header          DS                  likeds(tree_header_t) based(tree)
     D embeddedPtr     S               *
      /free
       embeddedPtr = rbtree_first(header.tree);
       return tree_rb_int_container_of(embeddedPtr);
      /end-free
     P                 E


     /**
      * \brief Get last element
      *
      * Returns the last element of the tree.
      *
      * \param Pointer to tree
      *
      * \return Pointer to user data structure or *null if the tree is empty.
      */
     P tree_rb_int_last...
     P                 B                   export
     D                 PI              *
     D   tree                          *   const
      *
     D header          DS                  likeds(tree_header_t) based(tree)
     D embeddedPtr     S               *
      /free
       embeddedPtr = rbtree_last(header.tree);
       return tree_rb_int_container_of(embeddedPtr);
      /end-free
     P                 E


     /**
      * \brief Get next element
      *
      * Returns the next element of the tree starting from the passed element.
      *
      * \param Pointer to user data structure
      *
      * \return Pointer to user data structure of the next node or *null if
      *         there are no more elements after the passed one.
      */
     P tree_rb_int_next...
     P                 B                   export
     D                 PI              *
     D   nodePtr                       *   const
      *
     D node            DS                  likeds(tree_rb_node_int_t)
     D                                     based(nodePtr)
     D embeddedPtr     S               *
      /free
       embeddedPtr = rbtree_next(%addr(node.node));
       return tree_rb_int_container_of(embeddedPtr);
      /end-free
     P                 E


     /**
      * \brief Get previous element
      *
      * Returns the previous element of the tree starting from the passed element.
      *
      * \param Pointer to user data structure
      *
      * \return Pointer to user data structure of the previous node or *null if
      *         there are no more elements before the passed one.
      */
     P tree_rb_int_previous...
     P                 B                   export
     D                 PI              *
     D   nodePtr                       *   const
      *
     D node            DS                  likeds(tree_rb_node_int_t)
     D                                     based(nodePtr)
     D embeddedPtr     S               *
      /free
       embeddedPtr = rbtree_previous(%addr(node.node));
       return tree_rb_int_container_of(embeddedPtr);
      /end-free
     P                 E


     /**
      * \brief Is empty
      *
      * Checks if the tree is empty.
      *
      * \param Pointer to tree
      *
      * \return *on = the tree is empty <br>
      *         *off = tree has elements
      */
     P tree_rb_isEmpty...
     P                 B                   export
     D                 PI              N
     D   tree                          *   const
      *
     D header          DS                  likeds(tree_header_t) based(tree)
     D tree_internal   DS                  likeds(rbtree_t) based(header.tree)
      /free
       return (tree_internal.root = *null);
      /end-free
     P                 E


     /**
      * \brief Get tree size
      *
      * Returns the number of elements/nodes in this tree.
      *
      * \param Pointer to tree
      *
      * \return Number of elements in the tree
      *
      * \info As the number of elements is not saved in the head data structure
      *       this procedure must traverse the whole tree before returning the
      *       result. This might take some time on huge trees.
      */
     P tree_rb_size...
     P                 B                   export
     D                 PI            10I 0
     D   tree                          *   const
      *
     D header          DS                  likeds(tree_header_t) based(tree)
     D tree_internal   DS                  likeds(rbtree_t) based(header.tree)
     D nodeCount       S             10I 0
      /free
       if (tree_internal.root <> *null);
         countNode(tree_internal.root : nodeCount);
       endif;

       return nodeCount;
      /end-free
     P                 E


     /**
      * \brief Contains key
      *
      * Checks if the tree contains an element with the passed key.
      *
      * \param Pointer to tree
      * \param Key
      *
      * \return *on = tree contains key <br>
      *         *off = tree does not contain key
      */
     P tree_rb_int_containsKey...
     P                 B                   export
     D                 PI              N
     D   tree                          *   const
     D   key                         10I 0 const
      *
     D header          DS                  likeds(tree_header_t) based(tree)
     D node            DS                  likeds(tree_rb_node_int_t)
     D                                     based(nodePtr)
     D embeddedPtr     S               *
      /free
       clear node;
       node.key = key;
       embeddedPtr = rbtree_lookup(%addr(node.node) : tree);
       return (embeddedPtr <> *null);
      /end-free
     P                 E


     /**
      * \brief Clear tree
      *
      * Removes all nodes from the tree recursively from top to bottom. All
      * memory allocated to the nodes will be freed.
      *
      * \param Pointer to tree
      *
      * \info This procedure traverses the whole tree to complete. The tree
      *       will also rebalance itself on every remove operation. This might
      *       be costly on huge trees. It might be simpler to build a whole new
      *       tree.
      */
     P tree_rb_clear...
     P                 B                   export
     D                 PI
     D   tree                          *   const
      *
     D header          DS                  likeds(tree_header_t) based(tree)
     D tree_internal   DS                  likeds(rbtree_t) based(header.tree)
      /free
       if (tree_internal.root <> *null);
         removeNodesRecursively(tree : tree_internal.root);
       endif;
      /end-free
     P                 E


     /**
      * \brief Remove node
      *
      * Removes a node from the tree. All allocated resources to the node will
      * be freed.
      *
      * \param Pointer to tree
      * \param Pointer to user data structure (node)
      */
     P tree_rb_int_removeByNode...
     P                 B                   export
     D                 PI
     D   tree                          *   const
     D   nodePtr                       * 
      *
     D header          DS                  likeds(tree_header_t) based(tree)
     D node            DS                  likeds(tree_rb_node_int_t)
     D                                     based(nodePtr)
      /free
       rbtree_remove(%addr(node.node) : header.tree);
       
       if (header.deallocate);
         cee_freeStorage(node.value : *omit);
         cee_freeStorage(nodePtr : *omit);
       endif;
      /end-free
     P                 E


     /**
      * \brief Remove node by key
      *
      * Removes a node from the tree. All allocated resources to the node will
      * be freed. The procedure returns normally if the key does not exist in
      * the tree.
      *
      * \param Pointer to tree
      * \param Key
      */
     P tree_rb_int_remove...
     P                 B                   export
     D                 PI
     D   tree                          *   const
     D   key                         10I 0 const
      *
     D header          DS                  likeds(tree_header_t) based(tree)
     D node            DS                  likeds(tree_rb_node_int_t)
     D embeddedPtr     S               *
     D userNode        DS                  likeds(tree_rb_node_int_t)
     D                                     based(userNodePtr)
      /free
       clear node;
       node.key = key;

       embeddedPtr = rbtree_lookup(%addr(node.node) : header.tree);
       if (embeddedPtr <> *null);
         userNodePtr = tree_rb_int_container_of(embeddedPtr);
         rbtree_remove(embeddedPtr : header.tree);

         if (header.deallocate);
           cee_freeStorage(userNode.value : *omit);
           cee_freeStorage(userNodePtr : *omit);
         endif;
       endif;
      /end-free
     P                 E


     /**
      * \brief Integer compare function
      *
      * This procedure is used for a tree with integer keys. It can be used
      * on the tree creation procedure like this:
      * <code>tree_rb_create(%paddr('tree_rb_int_compare'));</code>
      *
      * \param Embedded node pointer A
      * \param Embedded node pointer B
      *
      * \return less than 0 = key A is less than key B <br>
      *         0 = key A and B are equal <br>
      *         greater than 0 = key A is greater than key B
      */
     P tree_rb_int_compare...
     P                 B                   export
     D                 PI            10I 0
     D  a                              *   value
     D  b                              *   value
      *
     D nodeA           DS                  likeds(tree_rb_node_int_t)
     D                                     based(nodeAPtr)
     D nodeB           DS                  likeds(tree_rb_node_int_t)
     D                                     based(nodeBPtr)
     D offsets         DS                  likeds(tree_rb_node_int_t)
     D nodeOffset      S             10I 0

      /free
       nodeOffset = %addr(offsets.node) - %addr(offsets);
       nodeAPtr = a - nodeOffset;
       nodeBPtr = b - nodeOffset;

       return nodeA.key - nodeB.key;
      /end-free
     P                 E


     /**
      * \brief Return pointer to user data structure
      *
      * Returns a pointer to the user data structure of a node relative to the
      * passed embedded data structure pointer.
      *
      * \param Pointer to embedded data structure
      *
      * \return Pointer to user data structure or *null if input parameter is
      *         also *null
      */
     P tree_rb_int_container_of...
     P                 B                   export
     D                 PI              *
     D   eNode                         *   value
      *
     D offsets         DS                  likeds(tree_rb_node_int_t)
     D nodeOffset      S             10I 0
      /free
       if (eNode = *null);
         return *null;
       else;
         nodeOffset = %addr(offsets.node) - %addr(offsets);
         return eNode - nodeOffset;
       endif;
      /end-free
     P                 E


     P countNode       B
     D                 PI
     D   nodePtr                       *   const
     D   count                       10I 0
      *
     D node            DS                  likeds(rbnode_t) based(nodePtr)
      /free
       // count this node
       count += 1;

       if (node.left <> *null);
         countNode(node.left : count);
       endif;

       if (node.right <> *null);
         countNode(node.right : count);
       endif;
      /end-free
     P                 E

     P removeNodesRecursively...
     P                 B
     D                 PI
     D   tree                          *   const
     D   nodePtr                       *
      *
     D header          DS                  likeds(tree_header_t) based(tree)
     D userNode        DS                  likeds(tree_rb_node_int_t)
     D                                     based(userNodePtr)
     D node            DS                  likeds(rbnode_t) based(nodePtr)
      /free
       if (node.left <> *null);
         removeNodesRecursively(tree : node.left);
       endif;

       if (node.right <> *null);
         removeNodesRecursively(tree : node.right);
       endif;

       userNodePtr = tree_rb_int_container_of(nodePtr);
       rbtree_remove(nodePtr : header.tree);

       // dealloc memory
       if (header.deallocate);
         cee_freeStorage(userNode.value : *omit);
         cee_freeStorage(userNodePtr : *omit);
       endif;
      /end-free
     P                 E
